// app global css in SCSS form

// the container styles are thanks to https://stackoverflow.com/a/72361371
@import 'quasar.variables.scss';

$container-padding: 20px;

.container,
.container-sm,
.container-md,
.container-lg,
.container-xl {
  width: 100%;
  margin-right: auto;
  margin-left: auto;
}

@media (min-width: $breakpoint-sm-min) {
  .container,
  .container-sm {
    max-width: $breakpoint-xs - $container-padding;
  }
}

@media (min-width: $breakpoint-md-min) {
  .container,
  .container-sm,
  .container-md {
    max-width: $breakpoint-md - $container-padding;
  }
}

@media (min-width: $breakpoint-lg-min) {
  .container,
  .container-sm,
  .container-md,
  .container-lg {
    max-width: $breakpoint-lg - $container-padding;
  }
}
